# do not change this line manually
# this is managed by git tag and updated on every release
# NOTE: this represents the NEXT release version
__version__ = '0.1.5'

__default_tag_key__ = 'finetuner'

# define the high-level API: fit()
from typing import Dict, Optional, overload, TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from .helper import AnyDNN, DocumentArrayLike
    from .tuner.summary import Summary


# fit interface generated from Tuner
@overload
def fit(
    model: 'AnyDNN',  #: must be an embedding model
    train_data: 'DocumentArrayLike',
    eval_data: Optional['DocumentArrayLike'] = None,
    epochs: int = 10,
    batch_size: int = 256,
    loss: str = 'CosineSiameseLoss',
    learning_rate: float = 1e-3,
    optimizer: str = 'adam',
    optimizer_kwargs: Optional[Dict] = None,
    device: str = 'cpu',
) -> Tuple['AnyDNN', 'Summary']:
    ...


# fit interface derived from Tailor + Tuner
@overload
def fit(
    model: 'AnyDNN',
    train_data: 'DocumentArrayLike',
    eval_data: Optional['DocumentArrayLike'] = None,
    epochs: int = 10,
    batch_size: int = 256,
    loss: str = 'CosineSiameseLoss',
    learning_rate: float = 1e-3,
    optimizer: str = 'adam',
    optimizer_kwargs: Optional[Dict] = None,
    to_embedding_model: bool = True,  #: below are tailor args
    input_size: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
    layer_name: Optional[str] = None,
    output_dim: Optional[int] = None,
    freeze: bool = False,
    device: str = 'cpu',
) -> Tuple['AnyDNN', 'Summary']:
    ...


# fit interface from Labeler + Tuner
@overload
def fit(
    model: 'AnyDNN',  #: must be an embedding model
    train_data: 'DocumentArrayLike',
    interactive: bool = True,  #: below are labeler args
    clear_labels_on_start: bool = False,
    port_expose: Optional[int] = None,
    runtime_backend: str = 'thread',
    loss: str = 'CosineSiameseLoss',
    learning_rate: float = 1e-3,
    optimizer: str = 'adam',
    optimizer_kwargs: Optional[Dict] = None,
    device: str = 'cpu',
) -> Tuple['AnyDNN', None]:
    ...


# fit interface from Labeler + Tailor + Tuner
@overload
def fit(
    model: 'AnyDNN',
    train_data: 'DocumentArrayLike',
    interactive: bool = True,  #: below are labeler args
    clear_labels_on_start: bool = False,
    port_expose: Optional[int] = None,
    runtime_backend: str = 'thread',
    loss: str = 'CosineSiameseLoss',
    learning_rate: float = 1e-3,
    optimizer: str = 'adam',
    optimizer_kwargs: Optional[Dict] = None,
    to_embedding_model: bool = True,  #: below are tailor args
    input_size: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
    layer_name: Optional[str] = None,
    output_dim: Optional[int] = None,
    freeze: bool = False,
    device: str = 'cpu',
) -> Tuple['AnyDNN', None]:
    ...


def fit(
    model: 'AnyDNN', train_data: 'DocumentArrayLike', *args, **kwargs
) -> Tuple['AnyDNN', Optional['Summary']]:
    if kwargs.get('to_embedding_model', False):
        from .tailor import to_embedding_model

        model = to_embedding_model(model, *args, **kwargs)

    if kwargs.get('interactive', False):
        from .labeler import fit

        return model, fit(model, train_data, *args, **kwargs)
    else:
        from .tuner import fit

        return model, fit(model, train_data, *args, **kwargs)


# level them up to the top-level
from .tuner import save
from .tailor import display
from .embedding import embed



from collections import defaultdict
from copy import deepcopy
from math import ceil
from random import choices, sample, shuffle
from typing import Iterator, Sequence, Tuple


class RandomClassBatchSampler:
    """A batch sampler that fills the batch with an equal number of items from each
    class.

    It will try to make sure that all the items get used once in a single epoch, and
    that only ``num_items_per_class`` items in a batch come from one class. When there
    would not be enough items left from a single class to fill the batch, items will be
    randomly sampled from other already used items for this class.

    However, some cutoff might occur if there are not enough items left from different
    classes to fill the batch - in this case some items do not get used in that epoch.
    """

    def __init__(
        self, labels: Sequence[int], batch_size: int, num_items_per_class: int
    ):
        """Construct the batch sample.

        :param labels: A sequence of items labels, each label should be an integer
            denoting the class of the item
        :param batch_size: How many items to include in a batch
        :param num_items_per_class: How many items per class (unique labels) to include
            in a batch. For example, if ``batch_size`` is 20, and
            ``num_items_per_class`` is 4, the batch will consist of 4 items for each of
            the 5 classes.
        """
        if batch_size <= 0:
            raise ValueError('batch_size must be a positive integer')
        if num_items_per_class <= 0:
            raise ValueError('num_items_per_class must be a positive integer')

        if batch_size % num_items_per_class != 0:
            raise ValueError(
                'batch_size must be a multiple of num_items_per_class,'
                f' got batch_size={batch_size} and'
                f' num_items_per_class={num_items_per_class}'
            )

        self._num_items_per_class = num_items_per_class
        self._batch_size = batch_size
        self._num_classes = batch_size // num_items_per_class

        if len(set(labels)) < self._num_classes:
            raise ValueError(
                'Number of distinct classes in labels must be larger or equal than'
                'numbers of classes in a batch'
            )

        # Get mapping of labels (classes) and their positions
        self._class_to_labels = defaultdict(list)
        for idx, label in enumerate(labels):
            self._class_to_labels[label].append(idx)

        # Get class groups (ids to use in batches)
        self._cls_group_counts = []
        for key, val in self._class_to_labels.items():
            self._cls_group_counts.append([key, ceil(len(val) / num_items_per_class)])

        # Approximate (max) length
        self._length = ceil(
            sum([x[1] for x in self._cls_group_counts]) / self._num_classes
        )

    def __iter__(self) -> Iterator[Tuple[int, ...]]:

        # Shuffle class groups into batches, some cutoff may occur
        counts = deepcopy(self._cls_group_counts)
        group_batches = []
        while len(counts) >= self._num_classes:
            indices_batch = sample(range(len(counts)), self._num_classes)

            group_batch = [counts[ind][0] for ind in indices_batch]
            group_batches.append(group_batch)

            del_inds = []
            for ind in indices_batch:
                counts[ind][1] -= 1
                if counts[ind][1] == 0:
                    del_inds.append(ind)

            for ind in sorted(del_inds, reverse=True):
                del counts[ind]

        # Shuffle all labels within class, get extra samples to fill the batch
        class_to_labels = deepcopy(self._class_to_labels)
        for key in list(class_to_labels.keys()):
            key_labels = class_to_labels[key]

            missing_items = (
                self._num_items_per_class - len(key_labels) % self._num_items_per_class
            )
            if missing_items != self._num_items_per_class:
                key_labels += choices(key_labels, k=missing_items)
            shuffle(key_labels)

        # Construct batches
        batches = []
        for group_batch in group_batches:
            batch = []
            for cls_group in group_batch:
                cls_items = class_to_labels[cls_group][-self._num_items_per_class :]
                del class_to_labels[cls_group][-self._num_items_per_class :]
                batch += cls_items

            batches.append(batch)

        # Update length to actual one
        self._length = len(batches)

        return iter(batches)

    def __len__(self) -> int:
        """This gives the approximage length of the dataset before ``__iter__`` is
        called, and the actual length of the last produced list of batches after that.
        """
        return self._length


class SessionBatchSampler:
    """A batch sampler that fills the batch with items with items from as many sessions
    as possible.

    When constructing each batch, the sampler will start by adding all items of one
    session, and continue adding sessions in this way until the total number of items
    in equals ``batch_size`` - from the last session in batch only the first few items
    are taken, so as to not exceed the desider batch size.

    The last session in batch, if it was not included completely and was not the only
    session in the batch, will appear again in next the batch as the first added session
    (with all items, including those that appeared in the previous batch).

    It is assumed that the anchor document is the first document in its session -
    something that is always true if labels come from :class:`SessionDataset`
    """

    def __init__(
        self, labels: Sequence[Tuple[int, int]], batch_size: int, shuffle: bool = True
    ):
        """Constuct the batch sampler.

        :param labels: A sequence of items labels, each label should be a tuple with two
            integers - first one is the id of the session, while the second one denots
            the match type of the item (0 for root document, 1 for positive match
            and -1 for negative match)
        :param batch_size: How many items to include in a batch
        :param shuffle: Suffle the order of sessions. If false, will use the order of
            sessions which is given by ``labels``.
        """

        if batch_size <= 0:
            raise ValueError('batch_size must be a positive integer')

        self._labels = labels
        self._batch_size = batch_size
        self._shuffle = shuffle

        # The locations and types of labels for each session
        self._sessions = defaultdict(list)
        for ind, (session, match_type) in enumerate(labels):
            self._sessions[session].append([ind, match_type])

        # Approximate length
        self._length = ceil(len(labels) / self._batch_size)

    def __iter__(self):

        # Set up the order of sessions
        sessions = list(self._sessions.keys())
        if self._shuffle:
            shuffle(sessions)

        batches = []
        current_batch = []

        for session in sessions:
            session_items = self._sessions[session]
            session_exhausted = False

            while not session_exhausted:
                if len(current_batch) + len(session_items) <= self._batch_size:
                    current_batch += [x[0] for x in session_items]
                    session_exhausted = True
                else:
                    missing = self._batch_size - len(current_batch)

                    # If the session is larger than batch size, use and go to next
                    if len(current_batch) == 0:
                        session_exhausted = True

                    # Take the first one (root), sample the rest
                    current_batch += [session_items[0][0]]
                    current_batch += [
                        x[0] for x in sample(session_items[1:], missing - 1)
                    ]

                if len(current_batch) == self._batch_size:
                    batches.append(current_batch)
                    current_batch = []

        # Add remainder batch
        if len(current_batch):
            batches.append(current_batch)

        # Update length to actual one
        self._length = len(batches)

        return iter(batches)

    def __len__(self) -> int:
        """This gives the approximage length of the dataset before ``__iter__`` is
        called, and the actual length of the last produced list of batches after that.
        """
        return self._length



import abc
import warnings
from typing import (
    Generator,
    Optional,
    Union,
    Tuple,
    List,
    Dict,
)

from ..helper import AnyDNN, AnyTensor, AnyDataLoader, AnyOptimizer, DocumentArrayLike
from .summary import Summary


class BaseLoss:
    arity: int


class BaseTuner(abc.ABC):
    def __init__(
        self,
        embed_model: Optional[AnyDNN] = None,
        loss: Union[AnyDNN, str] = 'CosineSiameseLoss',
        **kwargs,
    ):
        """Create the tuner instance.

        :param embed_model: Model that produces embeddings from inputs
        :param loss: Either the loss object instance, or the name of the loss function.
            Currently available losses are ``CosineSiameseLoss``,
            ``EuclideanSiameseLoss``, ``EuclideanTripletLoss`` and ``CosineTripletLoss``
        """
        self._embed_model = embed_model
        self._loss = self._get_loss(loss)
        self._train_data_len = 0
        self._eval_data_len = 0

    def _get_optimizer_kwargs(self, optimizer: str, custom_kwargs: Optional[Dict]):
        """Merges user-provided optimizer kwargs with default ones."""

        DEFAULT_OPTIMIZER_KWARGS = {
            'adam': {'beta_1': 0.9, 'beta_2': 0.999, 'epsilon': 1e-08},
            'rmsprop': {
                'rho': 0.99,
                'momentum': 0.0,
                'epsilon': 1e-08,
                'centered': False,
            },
            'sgd': {'momentum': 0.0, 'nesterov': False},
        }

        try:
            opt_kwargs = DEFAULT_OPTIMIZER_KWARGS[optimizer]
        except KeyError:
            raise ValueError(
                f'Optimizer "{optimizer}" not supported, the supported'
                ' optimizers are "adam", "rmsprop" and "sgd"'
            )

        # Raise warning for non-existing keys passed
        custom_kwargs = custom_kwargs or {}
        extra_args = set(custom_kwargs.keys()) - set(opt_kwargs.keys())
        if extra_args:
            warnings.warn(
                f'The following arguments are not valid for the optimizer {optimizer}:'
                f' {extra_args}'
            )

        # Update only existing keys
        opt_kwargs.update((k, v) for k, v in custom_kwargs.items() if k in opt_kwargs)

        return opt_kwargs

    @property
    def embed_model(self) -> AnyDNN:
        """Get the base model of this object."""
        return self._embed_model

    @property
    def arity(self) -> int:
        """Get the arity of this object.

        For example,
            - ``arity = 2`` corresponds to the siamese network;
            - ``arity = 3`` corresponds to the triplet network.
        """
        return self._loss.arity

    @abc.abstractmethod
    def _get_optimizer(
        self, optimizer: str, optimizer_kwargs: Optional[dict], learning_rate: float
    ) -> AnyOptimizer:
        """Get the optimizer for training."""

    @abc.abstractmethod
    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        *args,
        **kwargs,
    ) -> Summary:
        """Fit the :py:attr:`.embed_model` on labeled data.

        Note that fitting changes the weights in :py:attr:`.embed_model` in-place. This allows one to consecutively
        call :py:func:`.fit` multiple times with different configs or data to get better models.
        """
        ...

    @abc.abstractmethod
    def save(self, *args, **kwargs):
        """Save the weights of the :py:attr:`.embed_model`."""
        ...

    @abc.abstractmethod
    def _get_loss(self, loss: Union[str, BaseLoss]) -> BaseLoss:
        """Get the loss layer."""
        ...

    @abc.abstractmethod
    def _get_data_loader(
        self, inputs: DocumentArrayLike, batch_size: int, shuffle: bool
    ) -> AnyDataLoader:
        """Get framework specific data loader from the input data."""
        ...

    @abc.abstractmethod
    def _train(
        self, data: AnyDataLoader, optimizer, description: str
    ) -> Tuple[List, List]:
        """Train the model on given labeled data"""
        ...

    @abc.abstractmethod
    def _eval(
        self, data: AnyDataLoader, description: str = 'Evaluating'
    ) -> Tuple[List, List]:
        """Evaluate the model on given labeled data"""
        ...


class BaseDataset:
    def __init__(
        self,
        inputs: DocumentArrayLike,
    ):
        super().__init__()
        self._inputs = inputs() if callable(inputs) else inputs


class BaseMiner(abc.ABC):
    @abc.abstractmethod
    def mine(
        self, embeddings: List[AnyTensor], labels: List[int]
    ) -> List[Tuple[int, ...]]:
        """Generate tuples/triplets from input embeddings and labels.

        :param embeddings: embeddings from model, should be a list of Tensor objects.
        :param labels: labels of each embeddings, embeddings with same label indicates same class.
        :return: tuple/triplet of label indices.
        """
        ...



import copy
import warnings
from collections import OrderedDict
from typing import Optional

import numpy as np
import paddle
from paddle import nn, Tensor

from ..base import BaseTailor
from ...helper import is_seq_int, LayerInfoType, AnyDNN


class PaddleTailor(BaseTailor):
    """Tailor class for Paddle DNN models.

    .. note::
        To use this class, you need to set ``input_size`` and ``input_dtype`` in :py:meth:`.__init__`
    """

    def summary(self, skip_identity_layer: bool = False) -> LayerInfoType:
        """Interpret the DNN model and produce model information.

        :param skip_identity_layer: If skip identity layer.
        :return: The model information stored as dict.
        """
        if not self._input_size:
            raise ValueError(
                f'{self.__class__} requires a valid `input_size`, but receiving {self._input_size}'
            )

        user_model = copy.deepcopy(self._model)
        dtypes = [self._input_dtype] * len(self._input_size)
        depth = len(list(user_model.sublayers()))
        for name, layer in user_model.named_sublayers():
            layer.name = name

        def _get_shape(output):
            if isinstance(output, (list, tuple)):
                output_shape = [_get_shape(o) for o in output]
                if len(output) == 1:
                    output_shape = output_shape[0]
            else:
                output_shape = list(output.shape)
            return output_shape

        def register_hook(layer):
            def hook(layer, input, output):

                class_name = str(layer.__class__).split('.')[-1].split("'")[0]

                layer_idx = len(summary)

                m_key = f'{class_name.lower()}_{layer_idx + 1}'

                summary[m_key] = OrderedDict()
                summary[m_key]['cls_name'] = layer.__class__.__name__
                summary[m_key]['name'] = m_key
                summary[m_key]['input_shape'] = _get_shape(input)
                summary[m_key]['output_shape'] = _get_shape(output)
                summary[m_key]['module_name'] = layer.name

                params = 0
                if paddle.in_dynamic_mode():
                    layer_state_dict = layer._parameters
                else:
                    layer_state_dict = layer.state_dict()

                for k, v in layer_state_dict.items():
                    params += np.prod(v.shape)

                summary[m_key]['nb_params'] = params
                summary[m_key]['trainable'] = any(
                    l.trainable for _, l in layer_state_dict.items()
                )

            if (
                not isinstance(layer, nn.Sequential)
                and not isinstance(layer, nn.LayerList)
                and (layer != user_model or depth < 1)
            ):
                hooks.append(layer.register_forward_post_hook(hook))
            # For rnn, gru and lstm layer
            elif hasattr(layer, 'could_use_cudnn') and layer.could_use_cudnn:
                hooks.append(layer.register_forward_post_hook(hook))

        x = [
            paddle.cast(paddle.rand([2, *in_size]), dtype)
            for in_size, dtype in zip(self._input_size, dtypes)
        ]

        # create properties
        summary = OrderedDict()
        hooks = []

        # register hook
        user_model.apply(register_hook)

        # make a forward pass
        user_model(*x)

        # remove these hooks
        for h in hooks:
            h.remove()

        results = []
        for idx, layer in enumerate(summary):
            output_shape = summary[layer]['output_shape']
            input_shape = summary[layer]['input_shape']
            is_embedding_layer = not (
                not output_shape
                or len(output_shape) != 2
                or not is_seq_int(output_shape)
                or summary[layer]['cls_name'] in self._model.__class__.__name__
            )

            if (
                skip_identity_layer
                and output_shape == input_shape
                and not summary[layer]['nb_params']
            ):
                # not an effective layer, often a wrapper/identity layer
                continue

            results.append(
                {
                    **summary[layer],
                    'output_features': output_shape[-1],
                    'output_shape_display': output_shape[1:],
                    'layer_idx': idx,
                    'is_embedding_layer': is_embedding_layer,
                }
            )
        return results

    def to_embedding_model(
        self,
        layer_name: Optional[str] = None,
        output_dim: Optional[int] = None,
        freeze: bool = False,
    ) -> AnyDNN:
        """Convert a general model from :py:attr:`.model` to an embedding model.

        :param layer_name: the name of the layer that is used for output embeddings. All layers *after* that layer
            will be removed. When set to ``None``, then the last layer listed in :py:attr:`.embedding_layers` will be used.
            To see all available names you can check ``name`` field of :py:attr:`.embedding_layers`.
        :param output_dim: the dimensionality of the embedding output.
        :param freeze: if set, then freeze all weights of the original model.
        :return: Converted embedding model.
        """
        model = copy.deepcopy(self._model)

        if layer_name:
            _all_embed_layers = {l['name']: l for l in self.embedding_layers}
            try:
                _embed_layer = _all_embed_layers[layer_name]
            except KeyError as e:
                raise KeyError(
                    f'`embedding_layer_name` must be one of {_all_embed_layers.keys()}, given {layer_name}'
                ) from e
        else:
            # when not given, using the last layer
            _embed_layer = self.embedding_layers[-1]

        if freeze:
            for param in model.parameters():
                param.trainable = False

        _relative_idx_to_embedding_layer = None
        _is_dense_layer_added = False
        for name, module in model.named_sublayers():
            if name == _embed_layer['module_name']:
                _relative_idx_to_embedding_layer = 0

                # corner-case
                if not output_dim and not layer_name:
                    for param in module.parameters():
                        param.trainable = True
                    else:
                        warnings.warn(
                            'The current configs results in a non-parametric model, '
                            'which is no trainable. '
                            'You may need to specify `output_dim` or `embedding_layer_name`.'
                        )

            if (
                _relative_idx_to_embedding_layer
                and _relative_idx_to_embedding_layer >= 1
            ):
                if _relative_idx_to_embedding_layer == 1 and output_dim:
                    replaced_layer = nn.Linear(
                        in_features=_embed_layer['output_features'],
                        out_features=output_dim,
                    )
                    _is_dense_layer_added = True
                else:
                    replaced_layer = _Identity()

                if (
                    '.' in name
                ):  # Note: in torchvision, nested layer names are named with '.' e.g. classifier.0
                    nested_module, layer = name.split('.')
                    setattr(getattr(model, nested_module), layer, replaced_layer)
                else:
                    setattr(model, name, replaced_layer)

            if _relative_idx_to_embedding_layer is not None:
                _relative_idx_to_embedding_layer += 1

        if output_dim and not _is_dense_layer_added:
            # the dense layer needs to be added after the last layer
            model = _LinearAtLast(
                model,
                in_features=_embed_layer['output_features'],
                out_features=output_dim,
            )

        return model


class _Identity(nn.Layer):
    """A placeholder identity operator that is argument-insensitive."""

    def forward(self, input_: Tensor) -> Tensor:
        return input_


class _LinearAtLast(nn.Layer):
    def __init__(self, model, *args, **kwargs):
        super().__init__()
        self._model = model
        self._linear = nn.Linear(*args, **kwargs)

    def forward(self, input_):
        return self._linear(self._model(input_))



import abc
import itertools
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar, Union

import numpy as np

from ... import __default_tag_key__
from ...helper import DocumentSequence


AnyLabel = TypeVar('AnyLabel')


class BaseDataset(abc.ABC, Generic[AnyLabel]):
    _labels: List[AnyLabel]

    @abc.abstractmethod
    def __getitem__(self, ind: int) -> Tuple[Union[np.ndarray, str], AnyLabel]:
        """
        Get the (preprocessed) content and label for the item at ``ind`` index in the
        dataset.
        """

    @property
    def labels(self) -> List[AnyLabel]:
        """ Get the list of labels for all items in the dataset."""
        return self._labels

    def __len__(self) -> int:
        return len(self._labels)


class ClassDataset(BaseDataset[int]):
    """ Dataset for enapsulating data where each item has a class label."""

    def __init__(
        self,
        docs: DocumentSequence,
        preprocess_fn: Optional[Callable[[Union[str, np.ndarray]], Any]] = None,
    ):
        """Create the :class:`ClassDataset` instance.

        :param docs: The documents for the dataset. Each document is expected to have
            - a content (only blob or text are accepted currently)
            - a class label, saved under ``tags['finetuner']['label']``. This class
              label should be an integer or a string
        :param preprocess_fn: A pre-processing function. It should take as input the
            content of an item in the dataset (currently only text of blob are
            accepted).
        """
        self._docs = docs
        self._preprocess_fn = preprocess_fn

        self._tag_labels_dict: Dict[Union[str, int], int] = {}
        self._labels: List[int] = []
        max_label = 0

        for doc in self._docs:
            try:
                tag = doc.tags[__default_tag_key__]['label']
            except KeyError as e:
                raise KeyError(
                    'The tag ["finetuner"]["label"] was not found in a document.'
                    ' When using ClassDataset all documents need this tag'
                ) from e
            label = self._tag_labels_dict.get(tag)

            if label is None:
                label = max_label
                self._tag_labels_dict[tag] = label
                max_label += 1

            self._labels.append(label)

    def __getitem__(self, ind: int) -> Tuple[Union[np.ndarray, str], int]:
        """
        Get the (preprocessed) content and label for the item at ``ind`` index in the
        dataset.
        """
        content = self._docs[ind].content
        label = self._labels[ind]

        if self._preprocess_fn:
            content = self._preprocess_fn(content)

        return (content, label)

    @property
    def labels(self) -> List[int]:
        """ Get the list of integer labels for all items in the dataset."""
        return self._labels


class SessionDataset(BaseDataset[Tuple[int, int]]):
    """Dataset for enapsulating data that comes in batches of "sessions".

    A session here is supposed to mean an anchor document, together with a set of
    matches, which may be either positive or negative inputs.
    """

    def __init__(
        self,
        docs: DocumentSequence,
        preprocess_fn: Optional[Callable[[Union[str, np.ndarray]], Any]] = None,
    ):
        """Create the :class:`SessionDataset` instance.

        :param docs: The documents for the dataset. Each document is expected to have
            - a content (only blob or text are accepted currently)
            - matches, which should also have content, as well a label, stored under
                ``tags['finetuner']['label']``, which be either 1 or -1, denoting
                whether the match is a positive or negative input in relation to the
                anchor document
        :param preprocess_fn: A pre-processing function. It should take as input the
            content of an item in the dataset (currently only text of blob are
            accepted).
        """
        self._docs = docs
        self._preprocess_fn = preprocess_fn

        self._locations: List[Tuple[int, int]] = []
        self._labels: List[Tuple[int, int]] = []

        num_docs = 0
        for i, doc in enumerate(self._docs):
            self._locations.append((i, -1))
            self._labels.append((i, 0))  # 0 is label for the anchor

            num_docs += 1

            for match_ind, match in enumerate(doc.matches):
                self._locations.append((i, match_ind))
                try:
                    tag = match.tags[__default_tag_key__]['label']
                except KeyError:
                    raise KeyError(
                        'The tag ["finetuner"]["label"] was not found in a document.'
                        ' When using ClassDataset all documents need this tag'
                    )

                self._labels.append((i, tag))
                num_docs += 1

    def __getitem__(self, ind: int) -> Tuple[Union[str, np.ndarray], Tuple[int, int]]:
        """
        Get the (preprocessed) content and label for the item at ``ind`` index in the
        dataset.
        """
        doc_ind, match_ind = self._locations[ind]

        if match_ind != -1:
            content = self._docs[doc_ind].matches[match_ind].content
        else:
            content = self._docs[doc_ind].content

        label = self._labels[ind]

        if self._preprocess_fn is not None:
            content = self._preprocess_fn(content)

        return (content, label)

    @property
    def labels(self) -> List[Tuple[int, int]]:
        """Get the list of labels for all items in the dataset.

        A label consists of two integers, the session ID (index of root document in
        original document array), and the type, which is 0 if the document is the
        anchor (root) document, 1 if it is a positive input (match), and -1 if it is
        a negative input (match)
        """
        return self._labels


class SiameseMixin:
    def __iter__(self):
        for d in self._inputs:
            d_blob = d.blob
            for m in d.matches:
                yield (d_blob, m.blob), np.float32(m.tags[__default_tag_key__]['label'])


class TripletMixin:
    def __iter__(self):
        for d in self._inputs:
            anchor = d.blob
            positives = []
            negatives = []
            for m in d.matches:
                if m.tags[__default_tag_key__]['label'] > 0:
                    positives.append(m.blob)
                else:
                    negatives.append(m.blob)

            for p, n in itertools.product(positives, negatives):
                yield (anchor, p, n), np.float32(0)



def get_dataset(module, arity: int):
    if arity == 2:

        return getattr(module, 'SiameseDataset')
    elif arity == 3:

        return getattr(module, 'TripletDataset')
    else:
        raise NotImplementedError



from typing import Dict, Optional, Union, List

import paddle
from jina.logging.profile import ProgressBar
from paddle.io import DataLoader
from paddle.optimizer import Optimizer

from . import losses, datasets
from ..base import BaseTuner, BaseLoss
from ..dataset.helper import get_dataset
from ..summary import ScalarSequence, Summary
from ...helper import DocumentArrayLike, AnyDataLoader


class PaddleTuner(BaseTuner):
    def _get_loss(self, loss: Union[BaseLoss, str]) -> BaseLoss:
        """Get the loss layer."""
        if isinstance(loss, str):
            return getattr(losses, loss)()
        elif isinstance(loss, BaseLoss):
            return loss

    def _get_data_loader(
        self, inputs: DocumentArrayLike, batch_size: int, shuffle: bool
    ) -> AnyDataLoader:
        """Get the paddle ``DataLoader`` from the input data."""
        ds = get_dataset(datasets, self.arity)
        return DataLoader(
            dataset=ds(inputs=inputs),
            batch_size=batch_size,
            shuffle=shuffle,
        )

    def _get_optimizer(
        self, optimizer: str, optimizer_kwargs: Optional[dict], learning_rate: float
    ) -> Optimizer:
        """Get the optimizer for training."""

        params = self._embed_model.parameters()
        optimizer_kwargs = self._get_optimizer_kwargs(optimizer, optimizer_kwargs)

        if optimizer == 'adam':
            return paddle.optimizer.Adam(
                parameters=params,
                learning_rate=learning_rate,
                beta1=optimizer_kwargs['beta_1'],
                beta2=optimizer_kwargs['beta_2'],
                epsilon=optimizer_kwargs['epsilon'],
            )
        elif optimizer == 'rmsprop':
            return paddle.optimizer.RMSProp(
                parameters=params, learning_rate=learning_rate, **optimizer_kwargs
            )
        elif optimizer == 'sgd':
            return paddle.optimizer.Momentum(
                parameters=params,
                learning_rate=learning_rate,
                momentum=optimizer_kwargs['momentum'],
                use_nesterov=optimizer_kwargs['nesterov'],
            )

    def _eval(
        self,
        data: AnyDataLoader,
        description: str = 'Evaluating',
        train_loss: Optional[ScalarSequence] = None,
    ) -> ScalarSequence:
        """Evaluate the model on given labeled data"""

        self._embed_model.eval()

        _summary = ScalarSequence('Eval Loss')

        with ProgressBar(
            description,
            message_on_done=lambda: f'{train_loss} | {_summary}',
            total_length=self._eval_data_len,
        ) as p:
            self._eval_data_len = 0
            for inputs, label in data:
                embeddings = [self._embed_model(inpt) for inpt in inputs]
                loss = self._loss(embeddings, label)

                _summary += loss.item()

                p.update(message=str(_summary))
                self._eval_data_len += 1

        return _summary

    def _train(
        self, data: AnyDataLoader, optimizer: Optimizer, description: str
    ) -> ScalarSequence:
        """Train the model on given labeled data"""

        self._embed_model.train()

        _summary = ScalarSequence('Train Loss')
        with ProgressBar(
            description,
            message_on_done=_summary.__str__,
            final_line_feed=False,
            total_length=self._train_data_len,
        ) as p:
            self._train_data_len = 0
            for inputs, label in data:
                # forward step
                embeddings = [self._embed_model(inpt) for inpt in inputs]
                loss = self._loss(embeddings, label)

                optimizer.clear_grad()

                loss.backward()
                optimizer.step()

                _summary += loss.item()

                p.update(message=str(_summary))
                self._train_data_len += 1
        return _summary

    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        optimizer: str = 'adam',
        optimizer_kwargs: Optional[Dict] = None,
        device: str = 'cpu',
        **kwargs,
    ) -> Summary:
        """Finetune the model on the training data.

        :param train_data: Data on which to train the model
        :param eval_data: Data on which to evaluate the model at the end of each epoch
        :param epochs: Number of epochs to train the model
        :param batch_size: The batch size to use for training and evaluation
        :param learning_rate: Learning rate to use in training
        :param optimizer: Which optimizer to use in training. Supported
            values/optimizers are:
            - ``"adam"`` for the Adam optimizer
            - ``"rmsprop"`` for the RMSProp optimizer
            - ``"sgd"`` for the SGD optimizer with momentum
        :param optimizer_kwargs: Keyword arguments to pass to the optimizer. The
            supported arguments, togethere with their defailt values, are:
            - ``"adam"``:  ``{'beta_1': 0.9, 'beta_2': 0.999, 'epsilon': 1e-08}``
            - ``"rmsprop"``::

                {
                    'rho': 0.99,
                    'momentum': 0.0,
                    'epsilon': 1e-08,
                    'centered': False,
                }

            - ``"sgd"``: ``{'momentum': 0.0, 'nesterov': False}``
        :param device: The device to which to move the model. Supported options are
            ``"cpu"`` and ``"cuda"`` (for GPU)
        """

        get_device(device)  #: this actually sets the device in Paddle

        _optimizer = self._get_optimizer(optimizer, optimizer_kwargs, learning_rate)

        m_train_loss = ScalarSequence('train')
        m_eval_loss = ScalarSequence('eval')

        for epoch in range(epochs):
            _data = self._get_data_loader(
                inputs=train_data, batch_size=batch_size, shuffle=False
            )
            lt = self._train(
                _data,
                _optimizer,
                description=f'Epoch {epoch + 1}/{epochs}',
            )
            m_train_loss += lt

            if eval_data:
                _data = self._get_data_loader(
                    inputs=eval_data, batch_size=batch_size, shuffle=False
                )

                le = self._eval(_data, train_loss=m_train_loss)
                m_eval_loss += le

        return Summary(m_train_loss, m_eval_loss)

    def save(self, *args, **kwargs):
        """Save the embedding model.

        You need to pass the path where to save the model in either ``args`` or
        ``kwargs`` (for ``path`` key).

        :param args: Arguments to pass to ``paddle.save`` function
        :param kwargs: Keyword arguments to pass to ``paddle.save`` function
        """
        paddle.save(self.embed_model.state_dict(), *args, **kwargs)


def get_device(device: str):
    """Get Paddle compute device.

    :param device: device name
    """

    # translate our own alias into framework-compatible ones
    if device == 'cuda':
        paddle.set_device('gpu:0')
    elif device == 'cpu':
        paddle.set_device('cpu')
    else:
        paddle.set_device(device)



import abc
from typing import (
    Optional,
    Tuple,
)

from ..helper import AnyDNN, LayerInfoType


class BaseTailor(abc.ABC):
    def __init__(
        self,
        model: AnyDNN,
        input_size: Optional[Tuple[int, ...]] = None,
        input_dtype: str = 'float32',
    ):
        """Tailor converts a general DNN model into an embedding model.

        :param model: a general DNN model
        :param input_size: a sequence of integers defining the shape of the input tensor. Note, batch size is *not* part
            of ``input_size``. It is required for :py:class:`PytorchTailor` and  :py:class:`PaddleTailor`, but not :py:class:`C`
        :param input_dtype: the data type of the input tensor.
        """
        self._model = model

        # multiple inputs to the network
        if isinstance(input_size, tuple):
            input_size = [input_size]

        self._input_size = input_size
        self._input_dtype = input_dtype

    @abc.abstractmethod
    def to_embedding_model(
        self,
        layer_name: Optional[str] = None,
        output_dim: Optional[int] = None,
        freeze: bool = False,
    ) -> AnyDNN:
        """Convert a general model from :py:attr:`.model` to an embedding model.

        :param layer_name: the name of the layer that is used for output embeddings. All layers *after* that layer
            will be removed. When set to ``None``, then the last layer listed in :py:attr:`.embedding_layers` will be used.
            To see all available names you can check ``name`` field of :py:attr:`.embedding_layers`.
        :param output_dim: the dimensionality of the embedding output.
        :param freeze: if set, then freeze all weights of the original model.

        """
        ...

    @property
    def embedding_layers(self) -> LayerInfoType:
        """Get all dense layers that can be used as embedding layer from the :py:attr:`.model`.

        :return: layers info as Dict.
        """
        _layers = self.summary()
        return [_l for _l in _layers if _l['is_embedding_layer']]

    @abc.abstractmethod
    def summary(self, include_identity_layer: bool = False) -> LayerInfoType:
        """The summary of the model architecture. To list all potential embedding layers, use :py:attr:`.embedding_layers`.

        :param include_identity_layer: if set, then identity layers are included and returned.
        :return: all layers info as Dict.
        """
        ...

    def display(self, *args, **kwargs) -> None:
        """Display the model architecture from :py:attr:`.summary` in a table.

        :param args: args pass to :py:attr:`.summary`
        :param kwargs: kwargs pass to :py:attr:`.summary`
        """
        from rich.table import Table
        from rich import print, box

        _summary = self.summary(*args, **kwargs)
        table = Table(box=box.SIMPLE)
        cols = ['name', 'output_shape_display', 'nb_params', 'trainable']
        for k in cols:
            table.add_column(k)
        for s in _summary:
            style = None
            if s['is_embedding_layer']:
                style = 'green'
            table.add_row(*map(str, (s[v] for v in cols)), style=style)
        print(
            table,
            '[green]Green[/green] layers can be used as embedding layers, '
            'whose [b]name[/b] can be used as [b]layer_name[/b] in to_embedding_model(...).',
        )



import tensorflow as tf
from tensorflow.keras.layers import Layer

from ..base import BaseLoss


class CosineSiameseLoss(BaseLoss, Layer):
    """Computes the loss for a siamese network using cosine distance.

    The loss for a pair of objects equals ``(target - cos_sim)^2``, where ``target``
    should equal 1 when both objects belong to the same class, and to -1 when they
    belong to different classes. The ``cos_sim`` represents the cosime similarity
    between both objects.

    The final loss is the average over losses for all pairs of objects in the batch.
    """

    arity = 2

    def call(self, inputs, **kwargs):
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the first objects of the pair
            - ``[N, D]`` tensor of embeddings of the second objects of the pair
            - ``[N, ]`` tensor of target values
        """

        l_emb, r_emb, target = inputs
        normalize_a = tf.nn.l2_normalize(l_emb, axis=-1)
        normalize_b = tf.nn.l2_normalize(r_emb, axis=-1)
        cos_similarity = tf.reduce_sum(tf.multiply(normalize_a, normalize_b), axis=-1)
        return tf.keras.losses.mse(target, cos_similarity)


class EuclideanSiameseLoss(BaseLoss, Layer):
    """Computes the loss for a siamese network using eculidean distance.

    This loss is also known as contrastive loss.

    The loss being optimized equals::

        [is_sim * dist + (1 - is_sim) * max(margin - dist, 0)]^2

    where ``target`` should equal 1 when both objects belong to the same class,
    and 0 otheriwse. The ``dist`` is the euclidean distance between the embeddings of
    the objects, and ``margin`` is some number, used here to ensure better stability
    of training.

    The final loss is the average over losses for all pairs of objects in the batch.
    """

    arity = 2

    def __init__(self, margin: float = 1.0):
        super().__init__()
        self.margin = margin

    def call(self, inputs, **kwargs):
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the first objects of the pair
            - ``[N, D]`` tensor of embeddings of the second objects of the pair
            - ``[N, ]`` tensor of target values
        """
        l_emb, r_emb, target = inputs
        eucl_dist = tf.reduce_sum(tf.math.squared_difference(l_emb, r_emb), axis=-1)
        is_similar = tf.cast(target > 0, tf.float32)

        loss = 0.5 * tf.square(
            is_similar * eucl_dist
            + (1 - is_similar) * tf.nn.relu(self.margin - eucl_dist)
        )
        return tf.reduce_mean(loss)


class EuclideanTripletLoss(BaseLoss, Layer):
    """Compute the loss for a triplet network using euclidean distance.

    The loss is computed as ``max(dist_pos - dist_neg + margin, 0)``, where ``dist_pos``
    is the euclidean distance between the anchor embedding and positive embedding,
    ``dist_neg`` is the euclidean distance between the anchor and negative embedding,
    and ``margin`` represents a wedge between the desired wedge between anchor-negative
    and anchor-positive distances.

    The final loss is the average over losses for all triplets in the batch.
    """

    arity = 3

    def __init__(self, margin: float = 1.0, **kwargs):
        super().__init__(**kwargs)
        self._margin = margin

    def call(self, inputs, **kwargs):
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the anchor objects
            - ``[N, D]`` tensor of embeddings of the positive objects
            - ``[N, D]`` tensor of embeddings of the negative objects
        """
        anchor, positive, negative, _ = inputs

        # Seems that tf.norm suffers from numeric instability as explained here
        # https://github.com/tensorflow/tensorflow/issues/12071
        dist_pos = tf.reduce_sum(tf.math.squared_difference(anchor, positive), axis=-1)
        dist_neg = tf.reduce_sum(tf.math.squared_difference(anchor, negative), axis=-1)

        dist_pos = tf.sqrt(tf.maximum(dist_pos, 1e-9))
        dist_neg = tf.sqrt(tf.maximum(dist_neg, 1e-9))

        return tf.reduce_mean(tf.nn.relu(dist_pos - dist_neg + self._margin))


class CosineTripletLoss(BaseLoss, Layer):
    """Compute the loss for a triplet network using cosine distance.

    The loss is computed as ``max(dist_pos - dist_neg + margin, 0)``, where ``dist_pos``
    is the cosine distance between the anchor embedding and positive embedding,
    ``dist_neg`` is the cosine distance between the anchor and negative embedding, and
    ``margin`` represents a wedge between the desired wedge between anchor-negative and
    anchor-positive distances.

    The final loss is the average over losses for all triplets in the batch.
    """

    arity = 3

    def __init__(self, margin: float = 1.0):
        super().__init__()
        self._margin = margin

    def call(self, inputs, **kwargs):
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the anchor objects
            - ``[N, D]`` tensor of embeddings of the positive objects
            - ``[N, D]`` tensor of embeddings of the negative objects
        """
        anchor, positive, negative, _ = inputs

        # Seems that tf.norm suffers from numeric instability as explained here
        # https://github.com/tensorflow/tensorflow/issues/12071
        normalize_a = tf.nn.l2_normalize(anchor, axis=-1)
        normalize_p = tf.nn.l2_normalize(positive, axis=-1)
        normalize_n = tf.nn.l2_normalize(negative, axis=-1)
        dist_pos = 1 - tf.reduce_sum(tf.multiply(normalize_a, normalize_p), axis=-1)
        dist_neg = 1 - tf.reduce_sum(tf.multiply(normalize_a, normalize_n), axis=-1)

        return tf.reduce_mean(tf.nn.relu(dist_pos - dist_neg + self._margin))



from typing import Optional

from tensorflow.keras import Model
from tensorflow.keras.layers import Dense

from ..base import BaseTailor
from ...helper import LayerInfoType, AnyDNN


class KerasTailor(BaseTailor):
    """Tailor class for Keras DNN models."""

    def summary(self, skip_identity_layer: bool = False) -> LayerInfoType:
        """Interpret the DNN model and produce model information.

        :param skip_identity_layer: If skip identity layer.
        :return: The model information stored as dict.
        """

        def _get_output_shape(layer):
            try:
                return layer.output_shape
            except:
                pass  #: return none when

        def _get_input_shape(layer):
            try:
                return layer.input_shape
            except:
                pass  #: return none when

        results = []
        for idx, layer in enumerate(self._model.layers):
            output_shape = _get_output_shape(layer)
            input_shape = _get_input_shape(layer)
            is_embedding_layer = not (
                not output_shape
                or len(output_shape) != 2
                or not isinstance(output_shape[-1], int)
            )

            if not layer.built and not getattr(layer, '_is_graph_network', False):
                # If a subclassed model has a layer that is not called in Model.call, the
                # layer will not be built and we cannot call layer.count_params().
                params = 0
            else:
                params = layer.count_params()

            if skip_identity_layer and output_shape == input_shape and not params:
                # not an effective layer, often a wrapper/identity layer
                continue

            results.append(
                {
                    'name': layer.name,
                    'cls_name': layer.__class__.__name__,
                    'input_shape': input_shape,
                    'output_shape': output_shape,
                    'output_shape_display': list(output_shape[1:]),
                    'output_features': output_shape[
                        -1
                    ],  #: this only makes sense when is_embedding_layer is True
                    'nb_params': params,
                    'layer_idx': idx,
                    'module_name': layer.name,  # duplicate as `name` to make different backends consistent
                    'is_embedding_layer': is_embedding_layer,
                    'trainable': layer.trainable if params else False,
                }
            )
        return results

    def to_embedding_model(
        self,
        layer_name: Optional[str] = None,
        output_dim: Optional[int] = None,
        freeze: bool = False,
    ) -> AnyDNN:

        """Convert a general model from :py:attr:`.model` to an embedding model.

        :param layer_name: the name of the layer that is used for output embeddings. All layers *after* that layer
            will be removed. When set to ``None``, then the last layer listed in :py:attr:`.embedding_layers` will be used.
            To see all available names you can check ``name`` field of :py:attr:`.embedding_layers`.
        :param output_dim: the dimensionality of the embedding output.
        :param freeze: if set, then freeze all weights of the original model.
        :return: Converted embedding model.
        """

        if layer_name:
            _all_embed_layers = {l['name']: l for l in self.embedding_layers}
            try:
                _embed_layer = _all_embed_layers[layer_name]
            except KeyError as e:
                raise KeyError(
                    f'`embedding_layer_name` must be one of {_all_embed_layers.keys()}, given {layer_name}'
                ) from e
        else:
            # when not given, using the last layer
            _embed_layer = self.embedding_layers[-1]

        index = _embed_layer['layer_idx']

        if output_dim:
            out = Dense(output_dim)(self._model.layers[index].output)
            model = Model(self._model.input, out)
        elif _embed_layer != self._model.layers[-1]:
            out = self._model.layers[index].output
            model = Model(self._model.input, out)
        else:
            model = self._model

        if freeze:
            for layer in model.layers:
                layer.trainable = False

        # the last layer must be trainable
        model.layers[-1].trainable = True
        return model



import numpy as np
from typing import List, Tuple
from itertools import combinations, groupby, permutations

from ..base import BaseMiner
from ...helper import AnyTensor


class SiameseMiner(BaseMiner):
    def mine(
        self, embeddings: AnyTensor, labels: List[int]
    ) -> List[Tuple[int, int, int]]:
        """Generate tuples from input embeddings and labels.

        :param embeddings: embeddings from model, should be a list of Tensor objects.
        :param labels: labels of each embeddings, embeddings with same label indicates same class.
        :return: a pair of label indices and their label as tuple.
        """
        assert len(embeddings) == len(labels)
        return [
            (left[0], right[0], 1) if left[1] == right[1] else (left[0], right[0], -1)
            for left, right in combinations(enumerate(labels), 2)
        ]


class TripletMiner(BaseMiner):
    def mine(
        self, embeddings: AnyTensor, labels: List[int]
    ) -> List[Tuple[int, int, int]]:
        """Generate triplets from input embeddings and labels.

        :param embeddings: embeddings from model, should be a list of Tensor objects.
        :param labels: labels of each embeddings, embeddings with same label indicates same class.
        :return: triplet of label indices follows the order of anchor, positive and negative.
        """
        assert len(embeddings) == len(labels)
        labels1 = np.expand_dims(labels, 1)
        labels2 = np.expand_dims(labels, 0)
        matches = (labels1 == labels2).astype(int)
        diffs = matches ^ 1
        np.fill_diagonal(matches, 0)
        triplets = np.expand_dims(matches, 2) * np.expand_dims(diffs, 1)
        idxes_anchor, idxes_pos, idxes_neg = np.where(triplets)
        return list(zip(idxes_anchor, idxes_pos, idxes_neg))


class SiameseSessionMiner(BaseMiner):
    def mine(
        self, embeddings: AnyTensor, labels: List[Tuple[int, int]]
    ) -> List[Tuple[int, int, int]]:
        """Generate tuples from input embeddings and labels.

        :param embeddings: embeddings from model, should be a list of Tensor objects.
        :param labels: labels of each embeddings, consist of session id and label.
          the labels are either -1 or 1.
        :return: a pair of label indices and their label as tuple.
        """
        assert len(embeddings) == len(labels)
        rv = []
        labels_with_index = [item + (index,) for index, item in enumerate(labels)]
        sorted_labels_with_index = sorted(labels_with_index, key=lambda x: x[0])
        for _, group in groupby(sorted_labels_with_index, lambda x: x[0]):
            _, session_labels, session_indices = zip(*group)
            for left, right in combinations(enumerate(session_labels), 2):
                if left[1] >= 0 and right[1] >= 0:
                    # 0 represents for anchor, 1 for positive, they form positive pairs.
                    rv.append((session_indices[left[0]], session_indices[right[0]], 1))
                elif (left[1] >= 0 and right[1] == -1) or (
                    left[1] == -1 and right[1] >= 0
                ):
                    # one of the item is positive or anchor, another one is negative
                    rv.append((session_indices[left[0]], session_indices[right[0]], -1))
                else:
                    # both are negatives
                    continue
        return rv


class TripletSessionMiner(BaseMiner):
    def mine(
        self, embeddings: AnyTensor, labels: List[Tuple[int, int]]
    ) -> List[Tuple[int, int, int]]:
        """Generate triplets from input embeddings and labels.

        :param embeddings: embeddings from model, should be a list of Tensor objects.
        :param labels: labels of each embeddings, consist of session id and label, the labels of
          anchor, positive, negative are 0, 1 and -1.
        :return: triplet of label indices follows the order of anchor, positive and negative.
        """
        assert len(embeddings) == len(labels)
        rv = []
        labels_with_index = [item + (index,) for index, item in enumerate(labels)]
        sorted_labels_with_index = sorted(labels_with_index, key=lambda x: x[0])
        for _, group in groupby(sorted_labels_with_index, lambda x: x[0]):
            anchor_pos_session_labels = []
            anchor_pos_session_indices = []
            neg_session_indices = []
            for _, session_label, session_index in group:
                if session_label >= 0:
                    anchor_pos_session_labels.append(session_label)
                    anchor_pos_session_indices.append(session_index)
                else:
                    neg_session_indices.append(session_index)
            for anchor, pos in permutations(enumerate(anchor_pos_session_labels), 2):
                for neg_idx in neg_session_indices:
                    rv.append(
                        (
                            anchor_pos_session_indices[anchor[0]],
                            anchor_pos_session_indices[pos[0]],
                            neg_idx,
                        )
                    )
        return rv



from typing import Optional, TYPE_CHECKING, Type, Dict

from ..helper import AnyDNN, DocumentArrayLike, get_framework

if TYPE_CHECKING:
    from .base import BaseTuner
    from .summary import Summary


def _get_tuner_class(dnn_model: AnyDNN) -> Type['BaseTuner']:
    f_type = get_framework(dnn_model)

    if f_type == 'keras':
        from .keras import KerasTuner

        return KerasTuner
    elif f_type == 'torch':
        from .pytorch import PytorchTuner

        return PytorchTuner
    elif f_type == 'paddle':
        from .paddle import PaddleTuner

        return PaddleTuner


def fit(
    embed_model: AnyDNN,
    train_data: DocumentArrayLike,
    eval_data: Optional[DocumentArrayLike] = None,
    epochs: int = 10,
    batch_size: int = 256,
    loss: str = 'CosineSiameseLoss',
    learning_rate: float = 1e-3,
    optimizer: str = 'adam',
    optimizer_kwargs: Optional[Dict] = None,
    device: str = 'cpu',
    **kwargs,
) -> 'Summary':
    """Finetune the model on the training data.

    :param embed_model: an embedding model
    :param train_data: Data on which to train the model
    :param eval_data: Data on which to evaluate the model at the end of each epoch
    :param epochs: Number of epochs to train the model
    :param batch_size: The batch size to use for training and evaluation
    :param loss: Which loss to use in training. Supported
        losses are:
        - ``CosineSiameseLoss`` for Siamese network with cosine distance
        - ``EuclideanSiameseLoss`` for Siamese network with eculidean distance
        - ``CosineTripletLoss`` for Triplet network with cosine distance
        - ``EuclideanTripletLoss`` for Triplet network with eculidean distance
    :param learning_rate: Learning rate to use in training
    :param optimizer: Which optimizer to use in training. Supported
        values/optimizers are:
        - ``"adam"`` for the Adam optimizer
        - ``"rmsprop"`` for the RMSProp optimizer
        - ``"sgd"`` for the SGD optimizer with momentum
    :param optimizer_kwargs: Keyword arguments to pass to the optimizer. The
        supported arguments, togethere with their defailt values, are:
        - ``"adam"``:  ``{'beta_1': 0.9, 'beta_2': 0.999, 'epsilon': 1e-08}``
        - ``"rmsprop"``::

            {
                'rho': 0.99,
                'momentum': 0.0,
                'epsilon': 1e-08,
                'centered': False,
            }

        - ``"sgd"``: ``{'momentum': 0.0, 'nesterov': False}``
    :param device: The device to which to move the model. Supported options are
        ``"cpu"`` and ``"cuda"`` (for GPU)
    """
    ft = _get_tuner_class(embed_model)

    return ft(embed_model, loss=loss).fit(
        train_data,
        eval_data,
        epochs=epochs,
        batch_size=batch_size,
        device=device,
        learning_rate=learning_rate,
        optimizer=optimizer,
        optimizer_kwargs=optimizer_kwargs,
    )


def save(embed_model: AnyDNN, model_path: str, *args, **kwargs) -> None:
    """Save the embedding model.

    :param embed_model: The embedding model to save
    :param model_path: Path to file/folder where to save the model
    :param args: Arguments to pass to framework-specific tuner's ``save`` method
    :param kwargs: Keyword arguments to pass to framework-specific tuner's ``save``
        method
    """
    ft = _get_tuner_class(embed_model)

    ft(embed_model).save(model_path, *args, **kwargs)



from typing import (
    TypeVar,
    Sequence,
    Iterator,
    Union,
    Callable,
    List,
    Dict,
    Any,
)

from jina import Document, DocumentArray, DocumentArrayMemmap

AnyDNN = TypeVar(
    'AnyDNN'
)  #: The type of any implementation of a Deep Neural Network object
AnyTensor = TypeVar(
    'AnyTensor'
)  #: The type of any implementation of an tensor for model tuning
AnyDataLoader = TypeVar(
    'AnyDataLoader'
)  #: The type of any implementation of a data loader
AnyOptimizer = TypeVar(
    'AnyOptimizer'
)  #: The type of any implementation of an optimizer for training the model
DocumentSequence = TypeVar(
    'DocumentSequence',
    Sequence[Document],
    DocumentArray,
    DocumentArrayMemmap,
    Iterator[Document],
)  #: The type of any sequence of Document
DocumentArrayLike = Union[
    DocumentSequence,
    Callable[..., DocumentSequence],
]  #: The type :py:data:`DocumentSequence` or a function that gives :py:data:`DocumentSequence`

LayerInfoType = List[
    Dict[str, Any]
]  #: The type of embedding layer information used in Tailor


def get_framework(dnn_model: AnyDNN) -> str:
    """Return the framework that enpowers a DNN model.

    .. note::
        This is not a solid implementation. It is based on ``__module__`` name,
        the key idea is to tell ``dnn_model`` without actually importing the
        framework.

    :param dnn_model: a DNN model
    :return: `keras`, `torch`, `paddle` or ValueError

    """
    if 'keras' in dnn_model.__module__:
        return 'keras'
    elif 'torch' in dnn_model.__module__:  # note: cover torch and torchvision
        return 'torch'
    elif 'paddle' in dnn_model.__module__:
        return 'paddle'
    else:
        raise ValueError(f'can not determine the backend of {dnn_model!r}')


def is_seq_int(tp) -> bool:
    """Return True if the input is a sequence of integers."""
    return tp and isinstance(tp, Sequence) and all(isinstance(p, int) for p in tp)



from torch.utils.data import IterableDataset

from ..base import BaseDataset
from ..dataset import SiameseMixin, TripletMixin


class SiameseDataset(SiameseMixin, BaseDataset, IterableDataset):
    ...


class TripletDataset(TripletMixin, BaseDataset, IterableDataset):
    ...



from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base import BaseLoss


class CosineSiameseLoss(BaseLoss, nn.Module):
    """Computes the loss for a siamese network using cosine distance.

    The loss for a pair of objects equals ``(target - cos_sim)^2``, where ``target``
    should equal 1 when both objects belong to the same class, and to -1 when they
    belong to different classes. The ``cos_sim`` represents the cosime similarity
    between both objects.

    The final loss is the average over losses for all pairs of objects in the batch.
    """

    arity = 2

    def forward(
        self, embeddings: List[torch.Tensor], target: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss.

        :param embeddings: Should be a list or a tuple containing two tensors:
            - ``[N, D]`` tensor of embeddings of the first objects of the pair
            - ``[N, D]`` tensor of embeddings of the second object of the pair
        :param target: A ``[N, ]`` tensor of target values
        """
        l_emb, r_emb = embeddings
        cos_sim = F.cosine_similarity(l_emb, r_emb)
        loss = F.mse_loss(cos_sim, target)
        return loss


class EuclideanSiameseLoss(BaseLoss, nn.Module):
    """Computes the loss for a siamese network using eculidean distance.

    This loss is also known as contrastive loss.

    The loss being optimized equals::

        [is_sim * dist + (1 - is_sim) * max(margin - dist, 0)]^2

    where ``target`` should equal 1 when both objects belong to the same class,
    and 0 otheriwse. The ``dist`` is the euclidean distance between the embeddings of
    the objects, and ``margin`` is some number, used here to ensure better stability
    of training.

    The final loss is the average over losses for all pairs of objects in the batch.
    """

    arity = 2

    def __init__(self, margin: float = 1.0):
        super().__init__()
        self.margin = margin

    def forward(
        self, embeddings: List[torch.Tensor], target: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the first objects of the pair
            - ``[N, D]`` tensor of embeddings of the second objects of the pair
        :param target: A ``[N, ]`` tensor of target values
        """
        l_emb, r_emb = embeddings
        eucl_dist = F.pairwise_distance(l_emb, r_emb, p=2)
        is_similar = (target > 0).to(torch.float32)

        loss = 0.5 * torch.square(
            is_similar * eucl_dist + (1 - is_similar) * F.relu(self.margin - eucl_dist)
        )
        return loss.mean()


class EuclideanTripletLoss(BaseLoss, nn.Module):
    """Compute the loss for a triplet network using euclidean distance.

    The loss is computed as ``max(dist_pos - dist_neg + margin, 0)``, where ``dist_pos``
    is the euclidean distance between the anchor embedding and positive embedding,
    ``dist_neg`` is the euclidean distance between the anchor and negative embedding,
    and ``margin`` represents a wedge between the desired wedge between anchor-negative
    and anchor-positive distances.

    The final loss is the average over losses for all triplets in the batch.
    """

    arity = 3

    def __init__(self, margin: float = 1.0):
        super().__init__()
        self._margin = margin

    def forward(
        self, embeddings: List[torch.Tensor], target: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the anchor objects
            - ``[N, D]`` tensor of embeddings of the positive objects
            - ``[N, D]`` tensor of embeddings of the negative objects
        """
        anchor, positive, negative = embeddings
        dist_pos = F.pairwise_distance(anchor, positive, p=2)
        dist_neg = F.pairwise_distance(anchor, negative, p=2)

        return torch.mean(F.relu(dist_pos - dist_neg + self._margin))


class CosineTripletLoss(EuclideanTripletLoss):
    """Compute the loss for a triplet network using cosine distance.

    The loss is computed as ``max(dist_pos - dist_neg + margin, 0)``, where ``dist_pos``
    is the cosine distance between the anchor embedding and positive embedding,
    ``dist_neg`` is the cosine distance between the anchor and negative embedding, and
    ``margin`` represents a wedge between the desired wedge between anchor-negative and
    anchor-positive distances.

    The final loss is the average over losses for all triplets in the batch.
    """

    def forward(
        self, embeddings: List[torch.Tensor], target: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss.

        :param inputs: Should be a list or a tuple containing three tensors:
            - ``[N, D]`` tensor of embeddings of the anchor objects
            - ``[N, D]`` tensor of embeddings of the positive objects
            - ``[N, D]`` tensor of embeddings of the negative objects
        """
        anchor, positive, negative = embeddings
        dist_pos = 1 - F.cosine_similarity(anchor, positive)
        dist_neg = 1 - F.cosine_similarity(anchor, negative)

        return torch.mean(F.relu(dist_pos - dist_neg + self._margin))



from typing import Union

from jina import DocumentArray, DocumentArrayMemmap

from .helper import AnyDNN, get_framework


def embed(
    docs: Union[DocumentArray, DocumentArrayMemmap],
    embed_model: AnyDNN,
    device: str = 'cpu',
    batch_size: int = 256,
) -> None:
    """Fill the embedding of Documents inplace by using `embed_model`

    :param docs: the Documents to be embedded
    :param embed_model: the embedding model written in Keras/Pytorch/Paddle
    :param device: the computational device for `embed_model`, can be either `cpu` or `cuda`.
    :param batch_size: number of Documents in a batch for embedding
    """
    fm = get_framework(embed_model)
    globals()[f'_set_embeddings_{fm}'](docs, embed_model, device, batch_size)


def _set_embeddings_keras(
    docs: Union[DocumentArray, DocumentArrayMemmap],
    embed_model: AnyDNN,
    device: str = 'cpu',
    batch_size: int = 256,
):
    from .tuner.keras import get_device

    device = get_device(device)
    with device:
        for b in docs.batch(batch_size):
            b.embeddings = embed_model(b.blobs, training=False).numpy()


def _set_embeddings_torch(
    docs: Union[DocumentArray, DocumentArrayMemmap],
    embed_model: AnyDNN,
    device: str = 'cpu',
    batch_size: int = 256,
):
    from .tuner.pytorch import get_device

    device = get_device(device)

    import torch

    embed_model = embed_model.to(device)
    is_training_before = embed_model.training
    embed_model.eval()
    with torch.inference_mode():
        for b in docs.batch(batch_size):
            tensor = torch.tensor(b.blobs, device=device)
            b.embeddings = embed_model(tensor).cpu().detach().numpy()
    if is_training_before:
        embed_model.train()


def _set_embeddings_paddle(
    docs: Union[DocumentArray, DocumentArrayMemmap],
    embed_model: AnyDNN,
    device: str = 'cpu',
    batch_size: int = 256,
):
    from .tuner.paddle import get_device

    get_device(device)

    import paddle

    is_training_before = embed_model.training
    embed_model.eval()
    for b in docs.batch(batch_size):
        b.embeddings = embed_model(paddle.Tensor(b.blobs)).numpy()
    if is_training_before:
        embed_model.train()



import abc
from typing import Dict, Optional

from jina import Executor, DocumentArray, requests, DocumentArrayMemmap
from jina.helper import cached_property

from ..embedding import embed
from ..tuner import fit, save


class FTExecutor(Executor):
    def __init__(
        self,
        dam_path: str,
        metric: str = 'cosine',
        loss: str = 'CosineSiameseLoss',
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._all_data = DocumentArrayMemmap(dam_path)
        self._metric = metric
        self._loss = loss

    @abc.abstractmethod
    def get_embed_model(self):
        ...

    @abc.abstractmethod
    def get_stop_event(self):
        ...

    @cached_property
    def _embed_model(self):
        return self.get_embed_model()

    @requests(on='/next')
    def embed(self, docs: DocumentArray, parameters: Dict, **kwargs):
        if not docs:
            return
        self._all_data.reload()
        _catalog = self._all_data.sample(
            min(len(self._all_data), int(parameters.get('sample_size', 1000)))
        )

        embed(docs, self._embed_model)
        embed(_catalog, self._embed_model)

        docs.match(
            _catalog,
            metric=self._metric,
            limit=int(parameters.get('topk', 10)),
            exclude_self=True,
        )
        for d in docs.traverse_flat(['r', 'm']):
            d.pop('blob', 'embedding')

    @requests(on='/fit')
    def fit(self, docs: DocumentArray, parameters: Dict, **kwargs):
        fit(
            self._embed_model,
            docs,
            epochs=int(parameters.get('epochs', 10)),
            loss=self._loss,
        )

    @requests(on='/save')
    def save(self, parameters: Dict, **kwargs):
        model_path = parameters.get('model_path', 'trained.model')
        save(self._embed_model, model_path)
        print(f'model is saved to {model_path}')

    @requests(on='/terminate')
    def terminate(self, **kwargs):
        self.get_stop_event().set()


class DataIterator(Executor):
    def __init__(
        self,
        dam_path: str,
        labeled_dam_path: Optional[str] = None,
        clear_labels_on_start: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._all_data = DocumentArrayMemmap(dam_path)
        if not labeled_dam_path:
            labeled_dam_path = dam_path + '/labeled'
        self._labeled_dam = DocumentArrayMemmap(labeled_dam_path)
        if clear_labels_on_start:
            self._labeled_dam.clear()

    @requests(on='/feed')
    def store_data(self, docs: DocumentArray, **kwargs):
        self._all_data.extend(docs)

    @requests(on='/next')
    def take_batch(self, parameters: Dict, **kwargs):
        st = int(parameters.get('start', 0))
        ed = int(parameters.get('end', 1))

        self._all_data.reload()
        return self._all_data[st:ed]

    @requests(on='/fit')
    def add_fit_data(self, docs: DocumentArray, **kwargs):
        for d in docs.traverse_flat(['r', 'm']):
            d.content = self._all_data[d.id].content
        self._labeled_dam.extend(docs)
        return self._labeled_dam



import os
import tempfile
import threading
import webbrowser
from typing import Optional

import jina.helper
from jina import Flow
from jina.logging.predefined import default_logger

from .executor import FTExecutor, DataIterator
from ..helper import AnyDNN, DocumentArrayLike


def fit(
    embed_model: AnyDNN,
    train_data: DocumentArrayLike,
    clear_labels_on_start: bool = False,
    port_expose: Optional[int] = None,
    runtime_backend: str = 'thread',
    loss: str = 'CosineSiameseLoss',
    **kwargs,
) -> None:
    """Fit the model in an interactive UI.

    :param embed_model: The embedding model to fine-tune
    :param train_data: Data on which to train the model
    :param clear_labels_on_start: If set True, will remove all labeled data.
    :param port_expose: The port to expose.
    :param runtime_backend: The parallel backend of the runtime inside the Pea, either ``thread`` or ``process``.
    :param loss: Which loss to use in training. Supported
        losses are:
        - ``CosineSiameseLoss`` for Siamese network with cosine distance
        - ``EuclideanSiameseLoss`` for Siamese network with eculidean distance
        - ``CosineTripletLoss`` for Triplet network with cosine distance
        - ``EuclideanTripletLoss`` for Triplet network with eculidean distance
    :param kwargs: Additional keyword arguments.
    """
    dam_path = tempfile.mkdtemp()
    stop_event = threading.Event()

    class MyExecutor(FTExecutor):
        def get_embed_model(self):
            return embed_model

        def get_stop_event(self):
            return stop_event

    f = (
        Flow(
            protocol='http',
            port_expose=port_expose,
            prefetch=1,
            runtime_backend=runtime_backend,
        )
        .add(
            uses=DataIterator,
            uses_with={
                'dam_path': dam_path,
                'clear_labels_on_start': clear_labels_on_start,
            },
        )
        .add(
            uses=MyExecutor,
            uses_with={
                'dam_path': dam_path,
                'loss': loss,
            },
        )
    )

    f.expose_endpoint('/next')  #: for allowing client to fetch for the next batch
    f.expose_endpoint('/fit')  #: for signaling the backend to fit on the labeled data
    f.expose_endpoint('/feed')  #: for signaling the backend to fit on the labeled data
    f.expose_endpoint(
        '/save'
    )  #: for signaling the backend to save the current state of the model
    f.expose_endpoint('/terminate')  #: for terminating the flow from frontend

    def extend_rest_function(app):
        """Allow FastAPI frontend to serve finetuner UI as a static webpage"""
        from fastapi.staticfiles import StaticFiles

        p = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'ui')
        app.mount('/finetuner', StaticFiles(directory=p, html=True), name='static2')
        return app

    jina.helper.extend_rest_interface = extend_rest_function

    global is_frontend_open
    is_frontend_open = False

    with f:

        def open_frontend_in_browser(req):
            global is_frontend_open
            if is_frontend_open:
                return
            url_html_path = f'http://localhost:{f.port_expose}/finetuner'
            try:
                webbrowser.open(url_html_path, new=2)
            except:
                pass  # intentional pass, browser support isn't cross-platform
            finally:
                default_logger.info(f'Finetuner is available at {url_html_path}')
                is_frontend_open = True

        # feed train data into the labeler flow
        f.post(
            '/feed',
            train_data,
            request_size=10,
            show_progress=True,
            on_done=open_frontend_in_browser,
        )
        f.block(stop_event=stop_event)



import json
from typing import List, Union, Optional, Dict

import numpy as np

NumericType = Union[
    int, float, complex, np.number
]  #: The type of numerics including numpy data type


class ScalarSequence:
    def __init__(self, name: str):
        """Create a record for storing a list of scalar values e.g. losses/metrics

        :param name: the name of that record
        """

        self.name = name
        self._record = []

    def __iadd__(self, other: Union[List[NumericType], float, 'ScalarSequence']):
        if isinstance(other, list):
            self._record += other
        elif isinstance(other, ScalarSequence):
            self._record += other._record
        elif isinstance(other, np.ndarray) and np.squeeze(other).ndim == 1:
            self._record += [v for v in np.squeeze(other)]
        else:
            self._record.append(other)
        return self

    def __str__(self):
        if self._record:
            return f'{self.name}: {np.mean([float(loss) for loss in self._record]):.2f}'
        else:
            return f'{self.name} has no record'

    def floats(self) -> List[NumericType]:
        """Return all numbers as a list of Python native float """
        return [float(v) for v in self._record]

    def __bool__(self):
        return bool(self._record)


class Summary:
    def __init__(self, *records: ScalarSequence):
        """Create a collection of summaries. """
        self._records = [r for r in records if r]

    def __iadd__(self, other: 'Summary'):
        if isinstance(other, Summary):
            self._records += other._records
        return self

    def save(self, filepath: str):
        """Store all summary into a JSON file"""
        with open(filepath, 'w') as fp:
            json.dump(
                self.dict(),
                fp,
            )

    def dict(self) -> Dict[str, List[NumericType]]:
        """Return all summaries as a Dictionary, where key is the name and value is the record"""
        return {r.name: r.floats() for r in self._records}

    def plot(
        self,
        output: Optional[str] = None,
        max_plot_points: Optional[int] = None,
        **kwargs,
    ):
        """Plot all records in the summary into one plot.

        .. note::
            This function requires ``matplotlib`` to be installed.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param max_plot_points: the maximum number of points to plot. When the actual number of plots is larger than
            given number, then a linspace sampling is conducted first to get the actual number of points for plotting.
        :param kwargs: extra kwargs pass to matplotlib.plot
        """
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(
            1,
            len(self._records),
            figsize=(6 * len(self._records), 6),
            constrained_layout=True,
        )
        if not isinstance(axes, np.ndarray):
            # when only one record, axes is not a list, so wrap it
            axes = [axes]

        plt_kwargs = dict(alpha=0.8, linewidth=1)
        plt_kwargs.update(kwargs)

        for idx, record in enumerate(self._records):
            axes[idx].plot(
                *self._sample_points(record.floats(), max_len=max_plot_points),
                **plt_kwargs,
            )
            axes[idx].set_ylabel(record.name)
            axes[idx].set_xlabel('Steps')

        if output:
            plt.savefig(output, bbox_inches='tight', pad_inches=0.1)
        else:
            plt.show()

    @staticmethod
    def _sample_points(arr, max_len: int):
        if not max_len or max_len > len(arr):
            return list(range(0, len(arr))), arr
        else:
            idx = np.round(np.linspace(0, len(arr) - 1, max_len)).astype(int)
            return idx, [arr[j] for j in idx]



from typing import Optional, Tuple, TYPE_CHECKING, Type

from ..helper import get_framework, AnyDNN

if TYPE_CHECKING:
    from .base import BaseTailor


def _get_tailor_class(dnn_model: AnyDNN) -> Type['BaseTailor']:
    f_type = get_framework(dnn_model)

    if f_type == 'keras':
        from .keras import KerasTailor

        return KerasTailor
    elif f_type == 'torch':
        from .pytorch import PytorchTailor

        return PytorchTailor
    elif f_type == 'paddle':
        from .paddle import PaddleTailor

        return PaddleTailor


def to_embedding_model(
    model: AnyDNN,
    layer_name: Optional[str] = None,
    output_dim: Optional[int] = None,
    freeze: bool = False,
    input_size: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
    **kwargs
) -> AnyDNN:
    """Convert a general model from :py:attr:`.model` to an embedding model.

    :param model: The DNN model to be converted.
    :param layer_name: the name of the layer that is used for output embeddings. All layers *after* that layer
        will be removed. When set to ``None``, then the last layer listed in :py:attr:`.embedding_layers` will be used.
        To see all available names you can check ``name`` field of :py:attr:`.embedding_layers`.
    :param output_dim: the dimensionality of the embedding output.
    :param freeze: if set, then freeze all weights of the original model.
    :param input_size: The input size of the DNN model.
    :param input_dtype: The input data type of the DNN model.
    """
    ft = _get_tailor_class(model)

    return ft(model, input_size, input_dtype).to_embedding_model(
        layer_name=layer_name, output_dim=output_dim, freeze=freeze
    )


def display(
    model: AnyDNN,
    input_size: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
) -> AnyDNN:
    """Display the model architecture from :py:attr:`.summary` in a table.

    :param model: The DNN model to display.
    :param input_size: The input size of the DNN model.
    :param input_dtype: The input data type of the DNN model.
    """
    ft = _get_tailor_class(model)

    return ft(model, input_size, input_dtype).display()



from typing import Dict, Optional, Union

import tensorflow as tf
from jina.logging.profile import ProgressBar
from tensorflow import keras
from tensorflow.keras.optimizers import Optimizer

from . import losses, datasets
from ..base import BaseTuner, BaseLoss
from ..dataset.helper import get_dataset
from ..summary import ScalarSequence, Summary
from ...helper import DocumentArrayLike, AnyDataLoader


class KerasTuner(BaseTuner):
    def _get_loss(self, loss: Union[BaseLoss, str]) -> BaseLoss:
        """Get the loss layer."""
        if isinstance(loss, str):
            return getattr(losses, loss)()
        elif isinstance(loss, BaseLoss):
            return loss

    def _get_data_loader(
        self, inputs: DocumentArrayLike, batch_size: int, shuffle: bool
    ) -> AnyDataLoader:
        """Get tensorflow ``Dataset`` from the input data."""

        ds = get_dataset(datasets, self.arity)
        input_shape = self.embed_model.input_shape[1:]

        tf_data = tf.data.Dataset.from_generator(
            lambda: ds(inputs),
            output_signature=(
                tuple(
                    tf.TensorSpec(shape=input_shape, dtype=tf.float32)
                    for _ in range(self.arity)
                ),
                tf.TensorSpec(shape=(), dtype=tf.float32),
            ),
        )

        if shuffle:
            tf_data = tf_data.shuffle(buffer_size=4096)

        return tf_data.batch(batch_size)

    def _get_optimizer(
        self, optimizer: str, optimizer_kwargs: Optional[dict], learning_rate: float
    ) -> Optimizer:
        """Get the optimizer for training."""

        optimizer_kwargs = self._get_optimizer_kwargs(optimizer, optimizer_kwargs)

        if optimizer == 'adam':
            return keras.optimizers.Adam(
                learning_rate=learning_rate, **optimizer_kwargs
            )
        elif optimizer == 'rmsprop':
            return keras.optimizers.RMSprop(
                learning_rate=learning_rate, **optimizer_kwargs
            )
        elif optimizer == 'sgd':
            return keras.optimizers.SGD(learning_rate=learning_rate, **optimizer_kwargs)

    def _train(
        self, data: AnyDataLoader, optimizer: Optimizer, description: str
    ) -> ScalarSequence:
        """Train the model on given labeled data"""

        _summary = ScalarSequence('Train Loss')
        with ProgressBar(
            description,
            message_on_done=_summary.__str__,
            final_line_feed=False,
            total_length=self._train_data_len,
        ) as p:
            self._train_data_len = 0
            for inputs, label in data:
                with tf.GradientTape() as tape:
                    embeddings = [self._embed_model(inpt) for inpt in inputs]
                    loss = self._loss([*embeddings, label])

                grads = tape.gradient(loss, self._embed_model.trainable_weights)
                optimizer.apply_gradients(
                    zip(grads, self._embed_model.trainable_weights)
                )

                _summary += loss.numpy()

                p.update(message=str(_summary))
                self._train_data_len += 1

        return _summary

    def _eval(
        self,
        data: AnyDataLoader,
        description: str = 'Evaluating',
        train_loss: Optional[ScalarSequence] = None,
    ) -> ScalarSequence:
        """Evaluate the model on given labeled data"""

        _summary = ScalarSequence('Eval Loss')

        with ProgressBar(
            description,
            message_on_done=lambda: f'{train_loss} | {_summary}',
            total_length=self._eval_data_len,
        ) as p:
            self._eval_data_len = 0
            for inputs, label in data:
                embeddings = [self._embed_model(inpt) for inpt in inputs]
                loss = self._loss([*embeddings, label])

                _summary += loss.numpy()

                p.update(message=str(_summary))
                self._eval_data_len += 1

        return _summary

    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        optimizer: str = 'adam',
        optimizer_kwargs: Optional[Dict] = None,
        device: str = 'cpu',
        **kwargs,
    ) -> Summary:
        """Finetune the model on the training data.

        :param train_data: Data on which to train the model
        :param eval_data: Data on which to evaluate the model at the end of each epoch
        :param epochs: Number of epochs to train the model
        :param batch_size: The batch size to use for training and evaluation
        :param learning_rate: Learning rate to use in training
        :param optimizer: Which optimizer to use in training. Supported
            values/optimizers are:
            - ``"adam"`` for the Adam optimizer
            - ``"rmsprop"`` for the RMSProp optimizer
            - ``"sgd"`` for the SGD optimizer with momentum
        :param optimizer_kwargs: Keyword arguments to pass to the optimizer. The
            supported arguments, togethere with their defailt values, are:
            - ``"adam"``:  ``{'beta_1': 0.9, 'beta_2': 0.999, 'epsilon': 1e-08}``
            - ``"rmsprop"``::

                {
                    'rho': 0.99,
                    'momentum': 0.0,
                    'epsilon': 1e-08,
                    'centered': False,
                }

            - ``"sgd"``: ``{'momentum': 0.0, 'nesterov': False}``
        :param device: The device to which to move the model. Supported options are
            ``"cpu"`` and ``"cuda"`` (for GPU)
        """

        _train_data = self._get_data_loader(
            inputs=train_data, batch_size=batch_size, shuffle=False
        )

        if eval_data:
            _eval_data = self._get_data_loader(
                inputs=eval_data, batch_size=batch_size, shuffle=False
            )

        _optimizer = self._get_optimizer(optimizer, optimizer_kwargs, learning_rate)

        m_train_loss = ScalarSequence('train')
        m_eval_loss = ScalarSequence('eval')

        with get_device(device):
            for epoch in range(epochs):
                lt = self._train(
                    _train_data,
                    _optimizer,
                    description=f'Epoch {epoch + 1}/{epochs}',
                )
                m_train_loss += lt

                if eval_data:
                    le = self._eval(_eval_data, train_loss=m_train_loss)
                    m_eval_loss += le

        return Summary(m_train_loss, m_eval_loss)

    def save(self, *args, **kwargs):
        """Save the embedding model.

        You need to pass the path where to save the model in either ``args`` or
        ``kwargs`` (for ``filepath`` key).

        :param args: Arguments to pass to ``save`` method of the embedding model
        :param kwargs: Keyword arguments to pass to ``save`` method of the embedding
            model
        """

        self.embed_model.save(*args, **kwargs)


def get_device(device: str):
    """Get tensorflow compute device.

    :param device: device name
    """

    # translate our own alias into framework-compatible ones
    if device == 'cuda':
        device = '/GPU:0'
    elif device == 'cpu':
        device = '/CPU:0'
    return tf.device(device)



import copy
import csv
import gzip
import os
import urllib.request
from collections import defaultdict
from pathlib import Path
from typing import Optional, Generator

import numpy as np
from jina import Document, DocumentArray
from jina.logging.profile import ProgressBar

from finetuner import __default_tag_key__


def _text_to_word_sequence(
    text, filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n', split=' '
):
    text = text.lower()

    translate_dict = {c: split for c in filters}
    translate_map = str.maketrans(translate_dict)
    text = text.translate(translate_map)

    seq = text.split(split)
    return [i for i in seq if i]


def _build_vocab(texts, min_freq=1):
    all_tokens = defaultdict(int)
    for text in texts:
        seq = _text_to_word_sequence(text)
        for s in seq:
            all_tokens[s] += 1

    # 0 for padding, 1 for unknown
    return {
        k: idx
        for idx, k in enumerate(
            (k for k, v in all_tokens.items() if v >= min_freq), start=2
        )
    }


def _text_to_int_sequence(text, vocab, max_len=None):
    seq = _text_to_word_sequence(text)
    vec = [vocab.get(s, 1) for s in seq]
    if max_len:
        if len(vec) < max_len:
            vec = [0] * (max_len - len(vec)) + vec
        elif len(vec) > max_len:
            vec = vec[-max_len:]
    return vec


def generate_qa_match(
    num_total: int = 481,
    num_neg: int = 0,
    pos_value: int = 1,
    neg_value: int = -1,
    to_ndarray: bool = True,
    max_seq_len: int = 100,
    is_testset: Optional[bool] = None,
) -> Generator[Document, None, None]:
    """Get a generator of QA data with synthetic negative matches.

    :param num_total: the total number of documents to return
    :param num_neg: the number of negative matches per document
    :param pos_value: the label value of the positive matches
    :param neg_value: the label value of the negative matches
    :param to_ndarray: if set, then `text` is tokenized into a fixed length `ndarray`
    :param max_seq_len: the maximum sequence length of each text.
    :param is_testset: If to generate test data, if set to None, will all data return
    :return:
    """
    num_doc = 0
    all_docs = DocumentArray(_download_qa_data(is_testset=is_testset))

    if to_ndarray:
        all_texts = (
            all_docs.get_attributes('tags__question')
            + all_docs.get_attributes('tags__answer')
            + all_docs.get_attributes('tags__wrong_answer')
        )
        vocab = _build_vocab(all_texts, min_freq=2)

    for d in all_docs:
        d.text = d.tags['question']
        m_p = Document(
            text=d.tags['answer'], tags={__default_tag_key__: {'label': pos_value}}
        )
        m_n = Document(
            text=d.tags['wrong_answer'],
            tags={__default_tag_key__: {'label': neg_value}},
        )
        if to_ndarray:
            d.blob = np.array(
                _text_to_int_sequence(d.text, vocab, max_seq_len), np.long
            )
            m_p.blob = np.array(
                _text_to_int_sequence(m_p.text, vocab, max_seq_len), np.long
            )
            m_n.blob = np.array(
                _text_to_int_sequence(m_n.text, vocab, max_seq_len), np.long
            )

        if num_neg > 0:
            d.matches.append(m_p)
            d.matches.append(m_n)
            cur_num_neg = 1
            if num_neg > 1:
                sampled_docs = all_docs.sample(num_neg)
                for n_d in sampled_docs:
                    if n_d.id != d.id:
                        new_nd = Document(
                            text=n_d.tags['answer'],
                            tags={__default_tag_key__: {'label': neg_value}},
                        )
                        if to_ndarray:
                            new_nd.blob = np.array(
                                _text_to_int_sequence(new_nd.text, vocab, max_seq_len),
                                np.long,
                            )
                        d.matches.append(new_nd)
                        cur_num_neg += 1
                        if cur_num_neg >= num_neg:
                            break
        num_doc += 1
        yield d

        if num_doc >= num_total:
            break


def generate_fashion_match(
    num_total: int = 60000,
    num_pos: int = 0,
    num_neg: int = 0,
    pos_value: int = 1,
    neg_value: int = -1,
    upsampling: int = 1,
    channels: int = 0,
    channel_axis: int = -1,
    is_testset: bool = False,
) -> Generator[Document, None, None]:
    """Get a Generator of fashion-mnist Documents with synthetic matches.

    :param num_total: the total number of documents to return
    :param num_pos: the number of positive matches per document
    :param num_neg: the number of negative matches per document
    :param pos_value: the label value of the positive matches
    :param neg_value: the label value of the negative matches
    :param upsampling: the rescale factor, must be integer and >=1. It rescales the image into a bigger image.
        For example, `upsampling=2` gives 56 x 56 images.
    :param channels: fashion-mnist data is gray-scale data, it does not have channel.
        One can set channel to 1 or 3 to simulate real grayscale or rgb imaga
    :param channel_axis: The axis for channels, e.g. for pytorch we expect B*C*W*H, channel axis should be 1.
    :param is_testset: If to generate test data
    :return:
    """
    _orginal_fashion_doc = _download_fashion_doc(
        upsampling=upsampling,
        channels=channels,
        channel_axis=channel_axis,
        is_testset=is_testset,
    )

    n_d = 0
    if num_pos > 0 or num_neg > 0:
        # need to build synthetic matches
        all_docs = DocumentArray(_orginal_fashion_doc)

        copy_all_docs = copy.deepcopy(all_docs)
        rv = copy_all_docs.split('class')

        for od in all_docs:
            pos_label = od.tags['class']
            pos_samples = rv[pos_label].sample(num_pos)
            for d in pos_samples:
                d.tags[__default_tag_key__] = {'label': pos_value}

            neg_samples = DocumentArray()
            while len(neg_samples) < num_neg:
                neg_samples.extend(
                    d
                    for d in copy_all_docs.sample(num_neg)
                    if d.tags['class'] != pos_label
                )
            neg_samples = neg_samples[:num_neg]

            for d in neg_samples:
                d.tags[__default_tag_key__] = {'label': neg_value}

            od.matches.extend(pos_samples)
            od.matches.extend(neg_samples)
            n_d += 1
            yield od
            if n_d >= num_total:
                break
    else:
        for d in _orginal_fashion_doc:
            n_d += 1
            yield d
            if n_d >= num_total:
                break


def _download_qa_data(
    download_proxy=None, is_testset: Optional[bool] = False, **kwargs
) -> Generator[Document, None, None]:
    download_dir = './data'
    Path(download_dir).mkdir(parents=True, exist_ok=True)

    targets = {
        'covid-csv': {
            'url': 'https://static.jina.ai/chatbot/dataset.csv',
            'filename': os.path.join(download_dir, 'dataset.csv'),
        }
    }

    opener = urllib.request.build_opener()
    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
    if download_proxy:
        proxy = urllib.request.ProxyHandler(
            {'http': download_proxy, 'https': download_proxy}
        )
        opener.add_handler(proxy)
    urllib.request.install_opener(opener)
    with ProgressBar('download chatbot data') as t:
        for k, v in targets.items():
            if not os.path.exists(v['filename']):
                urllib.request.urlretrieve(
                    v['url'], v['filename'], reporthook=lambda *x: t.update(0.01)
                )

        with open(targets['covid-csv']['filename']) as fp:
            lines = csv.DictReader(fp)
            for idx, value in enumerate(lines):
                if is_testset is None:
                    yield Document(value)
                elif is_testset:
                    if idx % 2 == 0:
                        yield Document(value)
                else:
                    if idx % 2 == 1:
                        yield Document(value)


def _download_fashion_doc(
    download_proxy=None, is_testset=False, **kwargs
) -> Generator[Document, None, None]:
    download_dir = './data'
    Path(download_dir).mkdir(parents=True, exist_ok=True)

    targets = {
        'index-labels': {
            'url': 'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz',
            'filename': os.path.join(download_dir, 'index-labels'),
        },
        'query-labels': {
            'url': 'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz',
            'filename': os.path.join(download_dir, 'query-labels'),
        },
        'index': {
            'url': 'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz',
            'filename': os.path.join(download_dir, 'index-original'),
        },
        'query': {
            'url': 'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz',
            'filename': os.path.join(download_dir, 'query-original'),
        },
    }

    opener = urllib.request.build_opener()
    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
    if download_proxy:
        proxy = urllib.request.ProxyHandler(
            {'http': download_proxy, 'https': download_proxy}
        )
        opener.add_handler(proxy)
    urllib.request.install_opener(opener)
    with ProgressBar('download fashion-mnist') as t:
        for k, v in targets.items():
            if not os.path.exists(v['filename']):
                urllib.request.urlretrieve(
                    v['url'], v['filename'], reporthook=lambda *x: t.update(0.01)
                )
            if k == 'index-labels' or k == 'query-labels':
                v['data'] = _load_labels(v['filename'])
            if k == 'index' or k == 'query':
                v['data'] = _load_mnist(v['filename'], **kwargs)
    if is_testset:
        partition = 'query'
    else:
        partition = 'index'
    for raw_img, lbl in zip(
        targets[partition]['data'], targets[f'{partition}-labels']['data']
    ):
        _d = Document(
            content=raw_img,
            tags={
                'class': int(lbl),
            },
        )
        _d.convert_image_blob_to_uri()
        _d.blob = (_d.blob / 255.0).astype(np.float32)
        yield _d


def _load_mnist(path, upsampling: int = 1, channels: int = 0, channel_axis=-1):
    """
    Load MNIST data

    :param path: path of data
    :param upsampling: the rescale factor, must be integer and >=1. It rescales the image into a bigger image.
        For example, upsampling=2 gives 56 x 56 images.
    :param channels: fashion-mnist data is gray-scale data, it does not have channel.
        One can set channel to 1 or 3 to simulate real grayscale or rgb imaga
    :param channel_axis: The axis for channels, e.g. for pytorch we expect B*C*W*H, channel axis should be 1.
    :return: MNIST data in np.array
    """
    upsampling_axes = [1, 2, 3]
    # remove B & C from axes, B has been excluded, only upsampling W & H
    if channel_axis < 0:
        channel_axis = upsampling_axes[channel_axis]
    upsampling_axes.remove(channel_axis)

    with gzip.open(path, 'rb') as fp:
        r = np.frombuffer(fp.read(), dtype=np.uint8, offset=16).reshape([-1, 28, 28])
        if channels > 0:
            r = np.stack((r,) * channels, axis=channel_axis)
        if upsampling > 1:
            r = r.repeat(upsampling, axis=upsampling_axes[0]).repeat(
                upsampling, axis=upsampling_axes[1]
            )
        return r


def _load_labels(path: str):
    """
    Load labels from path

    :param path: path of labels
    :return: labels in np.array
    """
    with gzip.open(path, 'rb') as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8, offset=8).reshape([-1, 1])
